// CIF amendments body
// $Id: CIF.C,v 1.9 2002/07/24 22:56:35 alexy Exp $

#include "CIF.h"
#include "Item.h"
#include "util.h"
#include "CorbaCif.h"

#include <algorithm>

const string CIF::atom = "_atom_id";
const string CIF::seq  = "_seq_id";
const string CIF::comp = "_comp_id";
const string CIF::asym = "_asym_id";
const string CIF::alt  = "_alt_id";

CIF::CIF()
{
  atomIndexParents.push_back("_atom_site.label_atom_id");
  atomIndexParents.push_back("_atom_site.auth_atom_id");

  seqIndexParents.push_back("_atom_site.label_seq_id");
  seqIndexParents.push_back("_atom_site.auth_seq_id");

  
  atomIndexSuffixes["_atom_site.label_seq_id"] = seq;
  atomIndexSuffixes["_atom_site.auth_seq_id"]  = seq;

  seqIndexSuffixes["_atom_site.label_comp_id"] = comp;
  seqIndexSuffixes["_atom_site.label_asym_id"] = asym;
  seqIndexSuffixes["_atom_site.label_alt_id"]  = alt;

  seqIndexSuffixes["_atom_site.auth_comp_id"]  = comp;
  seqIndexSuffixes["_atom_site.auth_asym_id"]  = asym;
  seqIndexSuffixes["_atom_site.auth_alt_id"]   = alt;

  atomIndexSuffixes.insert(seqIndexSuffixes.begin(),
			   seqIndexSuffixes.end());

}


bool CIF::isVector(const Item* ip) const
{
  return ip->isSubcategory("vector");
}


bool CIF::isCartesian(const Item* ip) const
{
  return ip->isSubcategory("cartesian_coordinate")
    || ip->isSubcategory("cartesian_coordinate_esd")
    || ip->isSubcategory("fractional_coordinate")
    || ip->isSubcategory("fractional_coordinate_esd");
}


bool CIF::isMatrix(const Item* ip) const
{
  return ip->isSubcategory("matrix");
}


string CIF::vectorName(const Item* ip) const
{
    size_t i = ip->name().find_first_of('[');
    size_t j = ip->name().find_last_of(']');

    if (i == string::npos || j == string::npos
	|| i > j) return ip->name();

    return ip->name().substr(0, i) + ip->name().substr(j + 1);
}


string CIF::cartesianName(const Item* ip) const
{
  size_t i = ip->name().find("cartn");
  if (i == string::npos) 
    i = ip->name().find("fract");

  if (i != string::npos) 
    return ip->name().substr(0, i + 5) + ip->name().substr(i + 7);

  return ip->name();
}


string CIF::matrixName(const Item* ip) const
{
    size_t i = ip->name().find_first_of('[');
    size_t j = ip->name().find_last_of(']');

    if (i == string::npos || j == string::npos
	|| i > j) return ip->name();

    return ip->name().substr(0, i) + ip->name().substr(j + 1);
}


bool CIF::isVectorHead(const Item *ip) const
{
  return isVector(ip) 
    && ip->name().find("[1]") != string::npos
      && ip->name().find("][") == string::npos;
}


bool CIF::isVectorTail(const Item *ip, const string name) const
{
  return isVector(ip) && !isVectorHead(ip) 
    && vectorName(ip) == name;
}


bool CIF::isCartesianHead(const Item *ip) const
{
    return isCartesian(ip)
      && (ip->name().find("cartn_x") != string::npos
	  || ip->name().find("fract_x") != string::npos);
}


bool CIF::isCartesianTail(const Item *ip, const string name) const
{
  return isCartesian(ip) && !isCartesianHead(ip) 
    && cartesianName(ip) == name;
}


bool CIF::isMatrixHead(const Item *ip) const
{
  return isMatrix(ip) 
    && ip->name().find("[1][1]") != string::npos;
}


bool CIF::isMatrixTail(const Item *ip, const string name) const
{
  return isMatrix(ip) && !isMatrixHead(ip) 
    && matrixName(ip) == name;
}


string CIF::indexName(const Item* ip) const
{
  if (util::hasSuffix(ip->name(), "_id"))
    return util::cutSuffix(ip->name(), "_id");
  else
    return ip->name();
}


bool   CIF::isAtomIndexHead(const Item *ip) const
{
//    for (int i = 0; i < ip->_parents.size(); ++i)
//      // or could use ip->hasParent() here the other way round
//      if (util::present(atomIndexParents, ip->_parents[i])
//  	|| util::present(atomIndexParents, ip->fullName())) // for AtomSite
//        return true;

  for (Parents::const_iterator
	 it = atomIndexParents.begin(); it != atomIndexParents.end(); ++it)
    if (ip->hasParent(*it) || ip->fullName() == *it) 
      {
	// cerr << "got AtomIndex head: " << ip->fullName() << endl;
	return true;
      }

  return false;
}


string CIF::atomIndexName(const Item *ip) const
{
  return util::cutSubstr(ip->name(), atom);
}


string   CIF::isAtomIndexTail(const Item *ip, const string name) const
{
  for (Suffixes::const_iterator
	 it = atomIndexSuffixes.begin();
       it != atomIndexSuffixes.end(); ++it) {
    if (ip->hasParent(it->first) 
	|| ip->fullName() == it->first ) { // for AtomSite
      string shortName = util::cutSubstr(ip->name(), it->second);
      if (shortName == name) return it->second;
    }
  }
  return "";
}


bool   CIF::isSeqIndexHead(const Item *ip) const
{
//    for (int i = 0; i < ip->_parents.size(); ++i)
//      if (util::present(seqIndexParents, ip->_parents[i]))
//        return true;

  for (Parents::const_iterator
	 it = seqIndexParents.begin(); it != seqIndexParents.end(); ++it)
    if (ip->hasParent(*it)) return true;

  return false;
}


string CIF::seqIndexName(const Item *ip) const
{
  return util::cutSubstr(ip->name(), seq);
}


string CIF::isSeqIndexTail(const Item *ip, const string name) const
{
  for (Suffixes::const_iterator
	 it = seqIndexSuffixes.begin();
       it != seqIndexSuffixes.end(); ++it) {
    if (ip->hasParent(it->first)) {
      string shortName = util::cutSubstr(ip->name(), it->second);
      if (shortName == name) return it->second;
    }
  }
  return "";
}
