// Category body
// $Id: Category.C,v 1.23 2003/05/04 17:53:06 alexy Exp $

#include "Category.h"
#include "global.h"
#include "IDL.h"
#include "options.h"
#include "util.h"
#include "db.h"
#include "Column.h"
#include "output.h"
#include "cxx.h"
#include "CorbaCifStruct.h" // enums for category classes

#include <algorithm>

Category::Category() {}

Category::Category(string name)
  : _name(name), _dbChecked(false), _dbAbsent(false)
{

  _nameDB = util::upper(_name); // just happens so in our DB

  // create the IDL type name out of CIF name
  // Capitalize each word and remove underscores

  _nameIDL = util::_2Title(_name);

  char * mandcode = _cb->GetCategoryMandatoryCode(name.c_str());
  _mandatory = mandcode && strcmp(mandcode, "yes") == 0; // NB: make case insensitive
  free(mandcode);

  _isStruct = !opt::valuetypes || global::ccif->categoryStruct(name);

  _groupIDL = global::ccif->categoryGroup(name);
}


Category::~Category()
{
  for (int i = 0; i < _items.size(); ++i) {
    delete _items[i];
  }
}


void Category::setCifBuilder(CifBuilder *cb)
{
  _cb = cb;
}


void Category::addItem(Item *ip)
{
  _items.push_back(ip);
}


string Category::name() const
{
  return _name;
}


string Category::nameIDL() const
{
  return _nameIDL;
}


const long Category::nItems() const
{
  return _items.size();
}


string Category::IDL(const short n)
{
  string r;

  r = IDL::tab(n) + (_isStruct ? "struct" : "valuetype") + ' ' + _nameIDL + '\n' +
      IDL::tab(n) + "{\n";

      if (!_isStruct) r += IDL::tab(n+1) + "factory create" + _nameIDL + "();\n\n";
       
    for (int i = 0; i < _items.size(); ++i) {
      const string& idl = _items[i]->IDL();
      if (idl != "") r += IDL::tab(n+1) + (_isStruct ? "" : "public ") + idl + ";\n";
      if (opt::more_info) r += _items[i]->moreInfo() + '\n';
    }
    r += IDL::tab(n) + "};\n";

    r += IDL::tab(n) + "typedef sequence<" + _nameIDL + "> "
      + _nameIDL + "List" + ";\n";
    return r;
}
    
  
CifBuilder *Category::_cb = 0;


void Category::aggregate()
{
  aggregateBlocks();
  aggregateTensors();
}


void Category::aggregateBlocks()
{
  // cerr << "CATEGORY aggregating block: " << _name << endl;
  
  string shortName;
  vector<Items::iterator> remove;
  Items add;
  int errCode;

  ccif::Blocks blocks = global::ccif->blocks(_name, errCode);
  if (errCode) return; // no blocks in this category

  for (ccif::Blocks::iterator 
	 it = blocks.begin(); it != blocks.end(); ++it) {

    // cerr << "Aggregating " << _name << ' ' << it->name << endl;
    
    ItemBlock *v = new ItemBlock(this, _name, it->name, scopedIDL(it->type));
    add.push_back(v);

    v->setMembers(global::ccif->blockItems(_name, it->name, errCode));

    if (errCode) {
      cerr << "In category " << _name << " error for block " << it->name << endl;
      continue;
    }

    
    Strings itemNames = global::ccif->blockItemNames(_name, it->name, errCode);

    for (Items::iterator 
	   jt = _items.begin(); jt != _items.end(); ++jt)
      if (util::present(itemNames, (*jt)->nameMember())) {
	util::erase(itemNames, (*jt)->nameMember());
	v->addChild(*jt);
	// cerr << "Adding " << (*jt)->name() << " under " << it->name << endl;
	// *jt = 0; NB: use auto_ptr when the Item* is in two lists now?
	remove.push_back(jt);

	v->linkMember(*jt);
      }

    if (!itemNames.empty()) {
      cerr << "In category " << _name << " block " << it->name << " has items left:";
      util::print_cerr(itemNames);
    }

    v->checkOptional();
    // v->printMembers();
  }

  sort(remove.begin(), remove.end());

  while (!remove.empty()) {
    _items.erase(remove.back());
    remove.pop_back();
  }

  _items.insert(_items.end(), add.begin(), add.end());

//   cerr << "Category " << _name << " has " << _items.size() << " items:" << endl;
//   for (Items::iterator 
// 	 it = _items.begin(); it != _items.end(); ++it)
//     cerr << '\t' << (*it)->name() << endl;
}


void Category::aggregateTensors()
{
  string shortName;
  vector<Items::iterator> remove;
  Items add;
  int errCode;

  ccif::Tensors tensors = global::ccif->tensors(_name, errCode);
  if (errCode) return; // no tensors in this category

  for (ccif::Tensors::iterator 
	 it = tensors.begin(); it != tensors.end(); ++it) {
    
    ItemTensor *v = new ItemTensor(this, _name, it->name, scopedIDL(it->type), it->row_major, it->kind);
    add.push_back(v);

    // cerr << "aggregating " << it->name << endl;
    
    v->setMembers(global::ccif->tensorItems(_name, it->name, errCode));

    if (errCode) {
      cerr << "In category " << _name << " error for tensor " << it->name << endl;
      continue;
    }
    
    Strings itemNames = global::ccif->tensorItemNames(_name, it->name, errCode);
    if (errCode) {
      cerr << "In category " << _name << " tensor " << it->name << " is empty!";
      continue;
    }

    // for (int i = 0; i < itemNames.size(); ++i) cerr << ' ' << itemNames[i]; cerr << endl;

    for (Items::iterator 
	   jt = _items.begin(); jt != _items.end(); ++jt)
      if (util::present(itemNames, (*jt)->name())) {
	util::erase(itemNames, (*jt)->name());
	v->addChild(*jt);
	// cerr << "Adding " << (*jt)->name() << " under " << it->name << endl;
	// *jt = 0; NB: use auto_ptr when the Item* is in two lists now?
	remove.push_back(jt);
	
	v->linkMember(*jt);
      }
    
    if (!itemNames.empty()) {
      cerr << "In category " << _name << " tensor " << it->name << " has items left:";
      util::print_cerr(itemNames);
    }

    v->checkOptional();
  }

  sort(remove.begin(), remove.end());

  while (!remove.empty()) {
    _items.erase(remove.back());
    remove.pop_back();
  }

  _items.insert(_items.end(), add.begin(), add.end());
}


void Category::index()
{
  string shortName;
  vector<Items::iterator> remove;
  Items add;

  for (Items::iterator 
	 it = _items.begin(); it != _items.end(); ++it)
      if ((*it)->hasParents()) {
	shortName = global::cif->indexName(*it);

	ItemIndex *v = new ItemIndex(*it, shortName, (*it)->name());
	add.push_back(v);

	remove.push_back(it);
      }

  sort(remove.begin(), remove.end());

  while (!remove.empty()) {
    _items.erase(remove.back());
    remove.pop_back();
  }

  _items.insert(_items.end(), add.begin(), add.end());
  
}


void Category::transform()
{
  checkOptional();
  index();
  aggregate(); // changes some IndexId to AtomIndex or SeqIndex!
}


bool Category::isMandatory() const
{
  return _mandatory;
}


bool Category::isOptional() const
{
  return !_mandatory;
}


void Category::numberFlags(int& n, const bool split)
{
  _optFlagName = "S_" + util::upper(_name); // not _nameIDL here?
  _optFlagNumber = n++;
  
  for (Items::iterator 
	 it = _items.begin(); it != _items.end(); ++it)
    (*it)->numberFlags(n, split);
}


string Category::optFlags(const short t)
{
  string r;

  r += IDL::tab(t) + IDL::flagType + ' ' + _optFlagName + " = " + util::str(_optFlagNumber) + ";\n";

  for (Items::iterator 
	 it = _items.begin(); it != _items.end(); ++it)
    r += (*it)->optFlags(t);
  
  r += '\n';
  return r;
}


void Category::checkOptional()
{
  _hasOptionalItems = false;

  for (Items::iterator 
	 it = _items.begin(); it != _items.end(); ++it)
    if ((*it)->isOptional()) {
      _hasOptionalItems = true;
      return;
    }
  return;
}



int Category::itemPosByName(const string name) const
{
  for (int i = 0; i < _items.size(); ++i)
    if (_items[i]->name() == name) return i;

  return -1;
}


Strings Category::itemNames() const
{
  Strings r;

  for (int i = 0; i < _items.size(); ++i)
    r.push_back(_items[i]->name());

  return r;
}


Item* Category::getItem(const string name) const
{
  int i = itemPosByName(name);

  if (i == -1) return 0;
  else return _items[i];
}


void Category::deleteItem(const string name)
{
  int i = itemPosByName(name);

  if (i == -1) return;
  else _items.erase(_items.begin() + i);
}
  

string Category::entryCalls(const short t)
{
  string r;

  r = IDL::tab(t) + "long " + _name + "_list_size()" + '\n'
    + IDL::tab(t+1) + "raises (DataAccessException);" + '\n';

  r += IDL::tab(t) + _nameIDL + "List " 
    + "get_" + _name + "_list()" + '\n'
    + IDL::tab(t+1) + "raises (DataAccessException);" + '\n';

  return r;
}


string DB2initialize(const string funName, const short t);
string DB2getConnection(const string funName, const short t);
string DB2execute(const string funName, const short t);
string DB2finish(const string funName, const short t);

string Category::DB2(const short t)
{
  if (!_dbChecked) dbCheck();
    
  string r, f, select, funName, funSizeName, funType;

  // ...list_size
  funType = "CORBA::Long";
  funSizeName = _name + "_list_size()";
  funName = "MMS_EntryImpl::" + funSizeName;

  f = IDL::tab(t) + funType + '\n' 
    + IDL::tab(t) + funName + '\n'
    + IDL::tab(t+1) + cxx::throws;
  
  // out::h << funType << endl << funName << endl <<
  //  '\t' << cxx::throws << ';' << endl << endl;
  if (opt::headers) out::h << f << ';' << endl << endl;
  
  r += f + '\n'
    +  IDL::tab(t) + "{\n";

  if (_dbAbsent) {
      r += IDL::tab(t+1) + "return 0;\n"
	  + IDL::tab(t) + "}\n\n";
      goto list; // spaghetti, here we go!
  }

  r += DB2initialize(funName, t);

  r += IDL::tab(t+1) + "SQLCHAR * stmt = ( SQLCHAR * )\n"
    +  IDL::tab(t+2) + "\"SELECT count(*) FROM " + db::DB::schema + '.' + _nameDB 
    + " WHERE structure_id = ?\";\n\n";

  r += DB2getConnection(funName, t);
  r += DB2execute(funName, t);

  // declare result structs

  r += IDL::tab(t+1) + "struct {\n"
    +  IDL::tab(t+2) + "SQLINTEGER ind;\n"
    +  IDL::tab(t+2) + "SQLINTEGER val;\n"
    +  IDL::tab(t+1) + "} total;\n\n";

  // bind the result column

  r += IDL::tab(t+1) + "sqlrc = SQLBindCol(hstmt, 1, SQL_C_LONG, &total.val, 0, &total.ind);\n"
    +  IDL::tab(t+1) + "if (sqlrc != SQL_SUCCESS)\n"
    +  IDL::tab(t+2) + "throw MMS::DataAccessException(\""
    + funName + "\", \"SQLBindCol\");\n\n";
  
  // fetch the count

  r += IDL::tab(t+1) + "sqlrc = SQLFetch(hstmt);\n"
    +  IDL::tab(t+1) + "if (sqlrc != SQL_SUCCESS && sqlrc != SQL_NO_DATA)\n"
    +  IDL::tab(t+2) + "throw MMS::DataAccessException(\""
    + funName + "\", \"SQLFetch\");\n\n";
  
  // if no data found

  r += IDL::tab(t+1) + "if (sqlrc == SQL_NO_DATA_FOUND)\n" 
    +  IDL::tab(t+2) + "throw MMS::DataAccessException(\""
    + funName + "\", \"no data\");\n\n"; // NB: really? or return 0?

  
  r += DB2finish(funName, t);

  r += IDL::tab(t+1) + "return total.val;\n"
    +  IDL::tab(t) + "}\n\n";



  // get...list
 list:
  funType = "MMS::" + _nameIDL + "List";
  funName = "MMS_EntryImpl::get_" + _name + "_list()";

  f = IDL::tab(t) + funType + '*' + '\n' // NB: '*'!
    + IDL::tab(t) + funName + '\n'
    + IDL::tab(t+1) + cxx::throws;
  
  // out::h << funType << '*' << endl << funName << endl << // NB: '*'!
  //  '\t' << cxx::throws << ';' << endl << endl;
  out::h << f << ';' << endl << endl;
  
  r += f + '\n'
    +  IDL::tab(t) + "{\n";

  if (_dbAbsent) {
      r += IDL::tab(t+1) + "return 0;\n"
	  + IDL::tab(t) + "}\n\n";
    return r;
  }

  r += DB2initialize(funName, t);

  // Initializations should not be crossed by the goto below...
  r += IDL::tab(t+1) + funType + "* listP = 0;\n" // in case of goto finish
    +  IDL::tab(t+1) + "CORBA::ULong count;\n"
    +  IDL::tab(t+1) + "SQLCHAR * stmt;\n\n";

  // we need connection to check list size...
  r += DB2getConnection(funName, t);
  
  // Check the list size, goto finish (to return the connection) at once if that's 0
  // Below, had "...listSize = " + funSizeName + ";\n" but it was consuming extra hdbc!
  r += IDL::tab(t+1) + "CORBA::ULong listSize = list_size(hdbc, \"" + _nameDB + "\", _id);\n"
    // after cd: "return <result>" is possible, but need to concoct the <result>,
    // plus rework the goto finish out of all other statements,
    // and first  provide a wrapper for stmt
    +  IDL::tab(t+1) + "if (!listSize) goto finish;\n\n;";
  

  // generate select
  db::Columns columns(1), someColumns; // first dummy column, as CLI is 1-based
    
  for (int i = 0; i < _items.size(); ++i) {
    someColumns = _items[i]->dbColumns();
    columns.insert(columns.end(), 
		   someColumns.begin(), someColumns.end());
  }

  // cerr << "We got " << columns.size() << " columns" << endl << flush;

  select = "SELECT";
  for (int j = 1; j < columns.size(); ++j)
    select +=  (j == 1 ? ' ' : ',') + columns[j].dbname;
  select += " FROM " + db::DB::schema + '.' + _nameDB + " WHERE structure_id = ?";

  r += IDL::tab(t+1) + "stmt = ( SQLCHAR * )\n"
    +  IDL::tab(t+2) + '"' + select + "\";\n\n";

  // prepare the statement

  global::db->prepareSQL(select);
  
  // Here, we shall study our columns...

  db::ColumnInfo ci;

  for (int i = 1; i < columns.size(); ++i) {
    global::db->describeColumn (i, ci);
    
    if (opt::debug)
      r += "column #" + util::str(i)
        + " has name " + (char*)ci.columnName 
        + " and length " + util::str(ci.columnSize)
        + " and type " + util::str(ci.dataType) + '\n';

    // name should be the uppercase version of dbname
    columns[i].name = (char*)ci.columnName;  // columns[i].name = util::upper(columns[i].name);
    
    columns[i].type = ci.dataType;
    columns[i].size = ci.columnSize;

    columns[i].ip->setSqlType(ci.dataType);
    columns[i].ip->setSqlSize(ci.columnSize);
  }

  global::db->finishSQL();

  if (opt::debug) return r;

  r += DB2execute(funName, t);

  // declare result structs
  for (int i = 1; i < columns.size(); ++i) {

    string type = global::db->cType(columns[i].type);

    r += IDL::tab(t+1) + "struct {\n"
      +  IDL::tab(t+2) + "SQLINTEGER ind;\n"
      +  IDL::tab(t+2) + type + " val";
    if (type == "SQLCHAR") r += '[' + util::str(columns[i].size) + ']';
    r += ";\n" + IDL::tab(t+1) + "} " + columns[i].name + ";\n\n";
  }

  // bind columns to variables

  for (int i = 1; i < columns.size(); ++i) {
    r += IDL::tab(t+1) + "sqlrc = SQLBindCol( hstmt, " 
      + util::str(i) + ", " + global::db->bType(columns[i].type) 
      + ", &" + columns[i].name + ".val, "
      + util::str(global::db->bSize(columns[i].type, columns[i].size)) 
      + ", &" + columns[i].name + ".ind);\n"
      + IDL::tab(t+1) + "if (sqlrc != SQL_SUCCESS)\n" 
      + IDL::tab(t+2) + "throw MMS::DataAccessException(\""
      + funName + "\", \"SQLBindCol " + util::str(i) + "\");\n\n";
  }

  // create the result list
  r += IDL::tab(t+1) + "listP = new " + funType + "(listSize);\n"
    +  IDL::tab(t+1) + "listP->length(listSize);\n"
    +  IDL::tab(t+1) + "count = 0;\n\n"
    +  IDL::tab(t+1) + "sqlrc = SQLFetch(hstmt);\n";

  // goto finish should be from above, based on listSize, i.e. never happen here:
  r += IDL::tab(t+1) + "if (sqlrc == SQL_NO_DATA_FOUND) goto finish;\n\n";

  r += IDL::tab(t+1) + "if (sqlrc != SQL_SUCCESS"
    " && sqlrc != SQL_SUCCESS_WITH_INFO" // investigate why is this necessary for our DB2 sometimes
    ")\n"
    +  IDL::tab(t+2) + "throw MMS::DataAccessException(\""
    + funName + "\", \"SQLFetch first\");\n\n";
    
  r += IDL::tab(t+1) +  "while (sqlrc != SQL_NO_DATA_FOUND) {\n";

  for (int i = 0; i < _items.size(); ++i) {
    r += _items[i]->DB2(t+2);
  
    // cerr << _name << ": "
    //	 << "processing item #" << i << ", "
    //	 << _items[i]->name() << endl;
  }
  r += '\n' 
    +  IDL::tab(t+2) + "++count;\n\n"
    +  IDL::tab(t+2) + "sqlrc = SQLFetch(hstmt);\n"
    +  IDL::tab(t+2) + "if (sqlrc != SQL_SUCCESS && sqlrc != SQL_NO_DATA_FOUND)\n"
    +  IDL::tab(t+3) + "throw MMS::DataAccessException(\""
    + funName + "\", \"SQLFetch loop\");\n"
    +  IDL::tab(t+1) + "}\n";


  r += DB2finish(funName, t);

  r += IDL::tab(t+1) + "return listP;\n"
    +  IDL::tab(t) + "}\n\n";

  return r;
}


string DB2initialize(const string funName, const short t)
{
  string r;
  
  r += IDL::tab(t+1) + "SQLRETURN   sqlrc = SQL_SUCCESS;\n"
    +  IDL::tab(t+1) + "SQLHANDLE   hdbc; // connection handle\n"
    +  IDL::tab(t+1) + "SQLHANDLE   hstmt ; // statement handle\n"
    +  IDL::tab(t+1) + "char        structure_id[10];\n\n";

  return r;
}


string DB2getConnection(const string funName, const short t)
{
  string r;

  r += IDL::tab(t+1) + "try {\n"

    // before cd:
    // replaced explicit connection get-release by wrapper con/des-structor:
    //    +  IDL::tab(t+2) + "hdbc = connpool::cp.getConnection();\n"
    +  IDL::tab(t+2) + "connpool::Connection conn;\n"
    +  IDL::tab(t+2) + "hdbc = conn.hdbc;\n"
    
    +  IDL::tab(t+2) + "} catch (DBException) {\n"
    +  IDL::tab(t+2) + "throw MMS::DataAccessException(\""
    + funName + "\", \"no connections!\");\n"
    +  IDL::tab(t+1) + "}\n\n";  

  // before cd:
  // r += IDL::tab(t+1) + "cerr << \"" + funName + " got connection\" << endl;\n";

  return r;
}


string DB2execute(const string funName, const short t)
{
  string r;
  
  // allocate statement handle -"-

  r += IDL::tab(t+1) + "sqlrc = SQLAllocHandle(SQL_HANDLE_STMT, hdbc, &hstmt);\n"
    +  IDL::tab(t+1) + "if (sqlrc != SQL_SUCCESS)\n"
    +  IDL::tab(t+2) + "throw MMS::DataAccessException(\""
    + funName + "\", \"SQLAllocHandle\");\n\n";

  // prepare the statement -"-

  r += IDL::tab(t+1) + "sqlrc = SQLPrepare(hstmt, stmt, SQL_NTS);\n"
    +  IDL::tab(t+1) + "if (sqlrc != SQL_SUCCESS)\n"
    +  IDL::tab(t+2) + "throw MMS::DataAccessException(\""
    + funName + "\", \"SQLPrepare\");\n\n";

  // bind the parameter -"-

  r += IDL::tab(t+1) + "sqlrc = SQLBindParameter(hstmt, 1, SQL_PARAM_INPUT, " 
    + "SQL_C_CHAR, SQL_CHAR, 10, 0, structure_id, 10, NULL);\n"
    +  IDL::tab(t+1) + "if (sqlrc != SQL_SUCCESS)\n"
    +  IDL::tab(t+2) + "throw MMS::DataAccessException(\""
    + funName + "\", \"SQLBindParameter\");\n\n";

  // execute the statement for the parameter -"-

  r += IDL::tab(t+1) + "strcpy( structure_id, (char*)_idV);\n"
    +  IDL::tab(t+1) + "sqlrc = SQLExecute(hstmt);\n"
    +  IDL::tab(t+1) + "if (sqlrc != SQL_SUCCESS)\n"
    +  IDL::tab(t+2) + "throw MMS::DataAccessException(\""
    + funName + "\", \"SQLExecute\");\n\n";

  return r;
}  


string DB2finish(const string funName, const short t)
{
  string r;
  
  r += IDL::tab(t+1) + "sqlrc = SQLFreeHandle(SQL_HANDLE_STMT, hstmt);\n"
    +  IDL::tab(t+1) + "if (sqlrc != SQL_SUCCESS)\n"
    +  IDL::tab(t+2) + "throw MMS::DataAccessException(\""
  + funName + "\", \"SQLFreeHandle\");\n\n";

  // return the connection to the pool and then the result -"-

  r += IDL::tab(t)   + "finish:\n";
  // before cd:
  // r += IDL::tab(t+1) + "connpool::cp.returnConnection(hdbc);\n";
  // r += IDL::tab(t+1) + "cerr << \"" + funName + " returned connection\" << endl;\n";

  return r;
}


void Category::dbCheck()
{
    if (_dbChecked) return;

    // cerr << "Category DB Check: " << _name << ' ';
    
    _dbChecked = true; // multiple returns

    _dbAbsent = !global::db->tablePresent(_nameDB);
    
    if (_dbAbsent) {
      // cerr << "ABSENT" << endl;
	for (Items::iterator
	   it = _items.begin(); it != _items.end(); ++it)
	  (*it)->setDbAbsent();
	return;
    }
    // cerr << "present";
    
    Strings columnNamesPresent = global::db->columns(/*table*/_nameDB);

    for (Items::iterator
	   it = _items.begin(); it != _items.end(); ++it) {
      // cerr << '\t' << (*it)->name() << ' ';
      (*it)->dbCheck(columnNamesPresent);
      // cerr << ((*it)->dbAbsent() ? "ABSENT" : "present") << endl;
    }
}


string Category::nameDB() const
{
  return _nameDB;
}


void Category::nameDB(const string name)
{
  _nameDB = name;
}
  
 
void Category::setFlags(Flags& flags, const string& id)
{
  if (!_dbChecked) dbCheck();

  cerr << "setFlags() called for category " << _name << endl;
  
  if (_dbAbsent || (_mandatory && !_hasOptionalItems)
      || !global::db->categoryPresent(_nameDB, id)) return;

  flags.set(_optFlagNumber);
  cerr << "Setting " << _optFlagName << '(' << _optFlagNumber << ')' << endl;
  
  for (Items::iterator
	 it = _items.begin(); it != _items.end(); ++it)
    (*it)->setFlags(flags, id);
}


string Category::scopedIDL(const string name) const
{
  //  if (_groupIDL == IDL::MMS_CLASS) return name;
  //  else return IDL::scopeMMS + "::" + name;
  return name;
}


IDL::CategoryGroup Category::groupIDL() const
{
  return _groupIDL;
}


string Category::client(const short n)
{
  string r;
  r += IDL::tab(n) + "// client call for category " + _name + ":\n";
  
  string  sizeType = "CORBA::Long"; // or just "long"
  string justTypeName = _nameIDL + "List";
  string justListName = util::uncap(justTypeName);
  string listNameV    = justListName + 'V';
  string justSizeName = justListName + "Size";

  string funSizeName = _name + "_list_size()";
  string funListName = "get_" + _name + "_list()";

  r += IDL::tab(n) + sizeType + ' ' + justSizeName
    + " = entryV->" + funSizeName + ";\n"
    + IDL::tab(n) + "cout << \"" + _name + " has \" << "
    + justSizeName + " << \" elements (\" << flush;\n\n"
    + IDL::tab(n) + "MMS::" + justTypeName + "_var " + listNameV
    + " = entryV->" + funListName + ";\n"
    + IDL::tab(n) + "cout << "
    + listNameV + "->length() << ')' << endl;\n\n";

  return r;
}
