#include "connpool.h"
#include <iostream> // cerr
using namespace std;

ConnectionPool::ConnectionPool()
  : _connected(false), _num(0)
{}

void
ConnectionPool::init(const char *dbAlias, 
		     const char *user, 
		     const char *pswd, 
		     const int  max)
  throw (DBException)
{
  _connected = false; _num = 0;
  
  pthread_mutex_init(&_mutex, NULL); // = PTHREAD_MUTEX_INITIALIZER;

  pthread_mutex_lock(&_mutex);

  _dbAlias = (char*)malloc(strlen(dbAlias)+1);
  strcpy(_dbAlias, dbAlias);
  _user = (char*)malloc(strlen(user)+1);
  strcpy(_user, user);
  _pswd = (char*)malloc(strlen(pswd)+1);
  strcpy(_pswd, pswd);

  SQLRETURN   sqlrc;

  sqlrc = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &_henv);
  if (sqlrc != SQL_SUCCESS) throw DBException();

  sqlrc = SQLSetEnvAttr(_henv, SQL_ATTR_CONNECTION_POOLING, (SQLPOINTER)SQL_CP_ONE_PER_HENV, 0);
  if (sqlrc != SQL_SUCCESS) throw DBException();

  createConnections(max);
  
  _connected = true;
  pthread_mutex_unlock(&_mutex);
}


ConnectionPool::ConnectionPool(const char *dbAlias, 
			       const char *user, 
			       const char *pswd, 
			       const int  max)
  throw (DBException)
{
  init(dbAlias, user, pswd, max);
}


ConnectionPool::~ConnectionPool()
{ 
  // When disconnectAll was in the destructor,
  // it threw DBException at us.
  // Now the user have to call disconnectAll() 
  // explicitly when done.
  // that happens if cp is a global, since the
  // other environment, including db2 related,
  // is possibly already deallocated.
  disconnectAll(); // for non-globals
}


void ConnectionPool::disconnectAll()
{

  if (!_connected) return;

  SQLRETURN   sqlrc;
  SQLHANDLE   hdbc;  /* connection handles */

  pthread_mutex_lock(&_mutex);

  while (!_pool.empty()) {
    hdbc = _pool.back();

    sqlrc = SQLDisconnect(hdbc);
    if (sqlrc != SQL_SUCCESS) throw DBException();  

    sqlrc = SQLFreeHandle(SQL_HANDLE_DBC, hdbc);
    if (sqlrc != SQL_SUCCESS) throw DBException();  

    _pool.pop_back();
    cerr << "deallocated connection number " << _pool.size()+1 << endl;
  }

  sqlrc = SQLFreeHandle(SQL_HANDLE_ENV, _henv);
  if (sqlrc != SQL_SUCCESS) throw DBException();  

  _connected = false;

  pthread_mutex_unlock(&_mutex);

}


SQLHANDLE ConnectionPool::getConnection()
  throw (DBException)
{
  pthread_mutex_lock(&_mutex);

  // cerr << "get connection";

  if (_pool.empty()) createConnections(1);

  SQLHANDLE   hdbc = _pool.back();

  _pool.pop_back();

  pthread_mutex_unlock(&_mutex);

  // cerr << ", have " << _pool.size() << endl;
  return hdbc;
}


void ConnectionPool::returnConnection(SQLHANDLE hdbc)
{
  pthread_mutex_lock(&_mutex);

  // cerr << "return connection";
  
  _pool.push_back(hdbc);

  pthread_mutex_unlock(&_mutex);

  // cerr << ", have " << _pool.size() << endl;
}



void ConnectionPool::createConnections(const int num)
{
  SQLRETURN   sqlrc;

  for (int i = 0; i < num; ++i) {
    SQLHANDLE hdbc;  /* connection handles */
    sqlrc = SQLAllocHandle(SQL_HANDLE_DBC, _henv, &hdbc);
    if (sqlrc != SQL_SUCCESS) throw DBException();
  
    sqlrc = SQLConnect( hdbc,
			(SQLCHAR *)_dbAlias, SQL_NTS,
			(SQLCHAR *)_user, SQL_NTS,
			(SQLCHAR *)_pswd, SQL_NTS
			) ;
    if (sqlrc != SQL_SUCCESS) throw DBException();

    sqlrc = SQLSetConnectAttr( hdbc,
			       SQL_ATTR_AUTOCOMMIT,
			       (SQLPOINTER)SQL_AUTOCOMMIT_ON, SQL_NTS) ;
    if (sqlrc != SQL_SUCCESS) throw DBException();  

    _pool.push_back(hdbc);
    cerr << "allocated connection number " << ++_num << endl;
  }
}
